package bvh.fry;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import robocode.ScannedRobotEvent;
import robocode.util.Utils;

public class Mjolnir  implements Constanten {
   public static boolean movementChallenge;
   public static boolean targettingChallenge;
   public boolean meleeGevecht;
   public Freya bot;
   public double xMax;
   public double yMax;
   public static Rectangle2D slagveld;
   public Ellipse2D.Double maximumCirkel;
   public boolean kanonUitlijningAfgerond = true;
   public Doel doel;
  /**
   * Mjolnir, or Thor's Hammer, means "lightning," and symbolized the God's power over Thunder and Lightning. 
   * The Hammer Mjolnir was said to always return after it had been thrown.
   */
  public Mjolnir(Freya bot, double xMax, double yMax) {
      this.bot = bot;
      this.xMax = xMax;
      this.yMax = yMax;

      this.maximumCirkel = new Ellipse2D.Double(0, 0, xMax, yMax);
      this.slagveld = new Rectangle2D.Double(0, 0, xMax, yMax);
      this.meleeGevecht = (bot.getOthers() > 1);

  }

   public void onScannedRobot(Doel d) {
      if (doel == null || !doel.actief
            || ((bot.getGunHeat() / bot.getGunCoolingRate()) > 5D
            && (bot.positie.distance(d) < 0.85 * bot.positie.distance(doel) || (d.energie < 20D && bot.positie.distance(d) < 300D)))) {
         doel = d;
      }

      vuurGolf(d);
   
   }

   public double[] geefGFStats(Doel d) {
      bot.log("geefGFStats():");

      double kogelSnelheid = BotUtils.berekenKogelSnelheid(bepaalKanonVuurkracht(d));
      long vliegtijd = (long) (bot.positie.distance(d) / kogelSnelheid);
      Point2D.Double doelPositie = d.schatPositie(bot.getTime() + vliegtijd);

      int slagIndex = bot.getOthers() == 1 ? 0 : 1;

      int positieIndex = slagveld.contains(doelPositie) ? 0 : (maximumCirkel.contains(doelPositie) ? 1 : 2);
      int afstandIndex = bepaalAfstandIndex(doelPositie);
      int lateraleSnelheidsIndex = (int) Math.abs(d.lateraleSnelheid);
      int transversaleSnelheidsIndex = (int) (8D - d.transversaleSnelheid);
      int versnellingsIndex = (int) Math.round(d.versnelling);

      if (versnellingsIndex != 0) {
         versnellingsIndex = (versnellingsIndex < 0) ? 1 : 2;
      }

      double[] gf = null;

      try {
         gf = d.guessFactors[slagIndex][positieIndex][afstandIndex][lateraleSnelheidsIndex][meleeGevecht ? 0
                                                                                                         : transversaleSnelheidsIndex][meleeGevecht
            ? 0 : versnellingsIndex];
      } catch (ArrayIndexOutOfBoundsException e) {
         System.out.println("geefGFStats(): ArrayIndexOutOfBoundsException ================");
         System.out.println("positieIndex               = " + positieIndex + " ( MAX= " + (POSITIESEGMENTEN - 1) + ")");
         System.out.println("afstandIndex               = " + afstandIndex + " ( MAX= " + (AFSTANDSEGMENTEN) + ")");
         System.out.println("snelheidsIndex             = " + lateraleSnelheidsIndex + " ( MAX= "
            + (LATERALESNELHEIDSSEGMENTEN - 1) + ")");
         System.out.println("transversaleSnelheidsIndex = " + transversaleSnelheidsIndex + " ( MAX= "
            + (TRANSVERSALESNELHEIDSSEGMENTEN - 1) + ")");
         System.out.println("geefGFStats(): einde ArrayIndexOutOfBoundsException ----------");
      }

      return (gf);
   }

   private int bepaalAfstandIndex(Point2D.Double d) {
      int index = 0;
      double afstand = bot.positie.distance(d);

      if (afstand > 150D) {
         afstand -= 150D;
         index = Math.min(1 + (int) Math.floor(afstand / 200D), AFSTANDSEGMENTEN);
      }

      return (index);
   }

   public void vuurGolf(Doel d) {
      bot.log("vuurGolf():");

      double kogelSnelheid = BotUtils.berekenKogelSnelheid(bepaalKanonVuurkracht(d));
      Point2D.Double doelpositie = new Point2D.Double(d.getX(), d.getY());

      Golf g = new Golf(bot.positie, doelpositie, kogelSnelheid, d.cirkelRichting, bot.getTime(), 0, geefGFStats(d));
      d.golven.add(g);
   }

   public int geefSchattingGF(Doel d, double kogelkracht) {
      bot.log("geefSchattingGF():");

      int schattingGF = GF_MIDDEN;

      // schiet direct op doel indien dit stationair is doordat het bijna is uitgeschakeld
      if (!(d.energie == 0 && d.snelheid == 0)) {
         // anders bepaal meest waarschijnlijke schietrichting
         double kogelSnelheid = BotUtils.berekenKogelSnelheid(kogelkracht);
         double[] stats = geefGFStats(d);
         double[] test = new double[(2 * GF_MIDDEN) + 1];
         int smoothWidth = Math.max((int) Math.round(
                  ((2D * GF_MIDDEN + 1D) * Math.atan((0.5 * BOTGROOTTE) / doel.distance(bot.positie))) / Math.asin(
                     MAXIMUM_SNELHEID / kogelSnelheid)), 1);

         // window de data met een driehoekig filter:
         for (int gf = 0; gf <= (2 * GF_MIDDEN); gf++) {
            for (int j = -smoothWidth; j < smoothWidth; j++) {
               double factor = (j == 0) ? 1D : (1 / Math.abs(j));
               int bin = BotUtils.maxMin(gf + j, 0, 2 * GF_MIDDEN);
               test[gf] += factor * stats[bin];
            }
         }

         // en bepaal beste GF:
         for (int gf = 0; gf <= (2 * GF_MIDDEN); gf++) {
            if (test[gf] > test[schattingGF]) {
               schattingGF = gf;
            }
         }
      }

      return (schattingGF);
   }

   public void uitvoeren() {
      bot.log("stuurKanon():");

      if (!movementChallenge) {
         Point2D.Double volgendePositie;
         
         if (bot.getOthers() > 1) {
            volgendePositie = new Point2D.Double(bot.getX(), bot.getY());
         }
         else {
            volgendePositie = bot.geefVolgendePositie();
         }
         double doelRichting = BotUtils.bepaalRichting(volgendePositie, doel);
      
         if (bot.getGunHeat() > 3D * bot.getGunCoolingRate()) {
            kanonUitlijningAfgerond = false;

            // richt kanon recht op doel:
            bot.setTurnGunRightRadians(Utils.normalRelativeAngle(doelRichting - bot.getGunHeadingRadians()));
         } else {
            double kanonVuurkracht = bepaalKanonVuurkracht(doel);

            if (kanonUitlijningAfgerond
                  && bot.getEnergy() > kanonVuurkracht
                  && Math.abs(bot.getGunTurnRemainingRadians()) < Math.atan2(18D, doel.distance(volgendePositie))) {
               kanonUitlijningAfgerond = false;

               if (bot.setFireBullet(kanonVuurkracht) != null) {
                  doel.schotenFreya++;
                  doel.gemKogelkrachtFreya += kanonVuurkracht;
               }
            } else {
               kanonUitlijningAfgerond = true;

               int bestGF = geefSchattingGF(doel, kanonVuurkracht);
               bot.setTurnGunRightRadians(Utils.normalRelativeAngle(doelRichting - bot.getGunHeadingRadians()
                     + doel.cirkelRichting * (bestGF - GF_MIDDEN)));
               bot.log("stuurKanon(): schattingGF = " + bestGF + " (doel = " + doel.naam + ")");
            }
         }
      }
   }

   public double bepaalKanonVuurkracht(Doel d) {
      double kanonVuurkracht;

      if (targettingChallenge) {
         kanonVuurkracht = MAXIMUMVUURKRACHT;
      } else {
         if (d.energie <= 16D) {
            kanonVuurkracht = BotUtils.berekenMinimumVuurkracht(d.energie);
         } else {
            int afstandIndex = bepaalAfstandIndex(d);
            kanonVuurkracht = Math.max(MAXIMUMVUURKRACHT - VUURKRACHTFACTOR * (Math.exp(afstandIndex / 1.5D)),
                  MINIMUMVUURKRACHT);
         }
      }

      return (kanonVuurkracht);
   }

}